!============================================================
! sdkLCS.clw -- List Column Sorter class implementation
!
!  2010 Simon Kemp, simon@sdk.demon.co.uk
!
!  This software is provided 'as-is', without any express or implied
!  warranty.  In no event will the author be held liable for any damages
!  arising from the use of this software.
!
!  Permission is granted to anyone to use this software for any purpose,
!  including commercial applications, and to alter it and redistribute it
!  freely, subject to the following restrictions:
!
!  1. The origin of this software must not be misrepresented; you must not
!     claim that you wrote the original software. If you use this software
!     in a product, an acknowledgment in the product documentation would be
!     appreciated but is not required.
!
!  2. Altered source versions must be plainly marked as such, and must not be
!     misrepresented as being the original software.
!
!  3. This notice may not be removed or altered from any source distribution.
!
!============================================================
sdkLCS.Init  procedure(long thelistfeq, *queue listqueue, long tweaklineheight=0)

    code

    free(self.columns)
    free(self.keys)
    self.alertkeystd = MouseLeftUp
    self.alertkeymulti = CtrlMouseLeft
    self.alertkeyclear = ShiftMouseLeft
    self.multikey = true
    self.prependindicators = false
    self.showcursoronsort = true
    self.sortonchanges = true
    self.listfeq = thelistfeq
    self.listfrom &= listqueue
    if tweaklineheight <> 0
      self.listfeq{PROP:LineHeight} = self.listfeq{PROP:LineHeight} + tweaklineheight
    end
    self.listfeq{PROP:Alrt, 255} = self.alertkeystd
    self.listfeq{PROP:Alrt, 255} = self.alertkeymulti
    self.listfeq{PROP:Alrt, 255} = self.alertkeyclear
    self.stickysort &= null
    clear(self.columns)
    loop
      self.columns.column += 1
      if ~self.listfeq{PROPLIST:Exists, self.columns.column}
        break
      end
      self.columns.headertext = self.listfeq{PROPLIST:Header, self.columns.column}
      self.columns.fieldno = self.listfeq{PROPLIST:FieldNo, self.columns.column}
      self.columns.columnwho = who(self.listfrom, self.columns.fieldno)
      self.columns.isastring = isstring(what(self.listfrom, self.columns.fieldno))
      add(self.columns)
    end
    self.blockcolumnoptions = false


sdkLCS.Kill  procedure

    code

    free(self.keys)
    self.ShowHeaders
    free(self.columns)
    self.listfrom &= null


sdkLCS.SetColumnOptions  procedure(<string nosorting>, <string casesensitive>, <string restrictsort>)

    code

    if self.blockcolumnoptions = true
      return
    end
    self.blockcolumnoptions = true
    loop self.columns.column = 1 to len(clip(nosorting))
      get(self.columns, +self.columns.column)
      if errorcode()
        cycle
      end
      if nosorting[self.columns.column] = true
        self.columns.sortoptions += 2^0 ! no sorting
        put(self.columns)
      end
    end
    loop self.columns.column = 1 to len(clip(casesensitive))
      get(self.columns, +self.columns.column)
      if errorcode()
        cycle
      end
      if casesensitive[self.columns.column] = true
        self.columns.sortoptions += 2^1 ! case-sensitive
        put(self.columns)
      end
    end
    loop self.columns.column = 1 to len(clip(restrictsort))
      get(self.columns, +self.columns.column)
      if errorcode()
        cycle
      end
      if inrange(restrictsort[self.columns.column],1,3)
        execute restrictsort[self.columns.column]
          self.columns.sortoptions += 2^2 ! restrict +
          self.columns.sortoptions += 2^3 ! restrict -
          begin
            if ~band(self.columns.sortoptions, 2^0)
              self.columns.sortoptions += 2^0 ! no sorting
            end
          end
        end
        put(self.columns)
      end
    end


sdkLCS.SetInitialSort  procedure(<*string variablesort>, <string initialsort>, long locked=0)

    code

    if ~omitted(variablesort)
      self.stickysort &= variablesort
      self.initialsort = left(variablesort)
    else
      self.initialsort = left(initialsort)
    end
    self.locked = locked


sdkLCS.Start procedure

myPtr1          long
myPtr2          long

    code

    self.blockcolumnoptions = true
    if self.initialsort
      case self.initialsort[1]
      of '+' orof '-'
      else
        self.initialsort = '+' & self.initialsort
      end
      myPtr1 = 1
      loop while myPtr1 < len(clip(self.initialsort))
        loop myPtr2 = myPtr1 + 1 to len(clip(self.initialsort)) + 1
          clear(self.keys)
          if ~inrange(val(self.initialsort[myPtr2]),val('0'),val('9'))
            self.keys.level = records(self.keys) + 1
            self.keys.column = deformat(self.initialsort[myPtr1 : myPtr2 - 1])
            if self.keys.column < 0
              self.keys.direction = -1
            else
              self.keys.direction = +1
            end
            self.keys.column = abs(self.keys.column)
            self.columns.column = self.keys.column
            get(self.columns,+self.columns.column)
            if ~errorcode() and ~band(self.columns.sortoptions,2^0) and ~(band(self.columns.sortoptions,2^2) and self.keys.direction < 0) and ~(band(self.columns.sortoptions,2^3) and self.keys.direction > 0)
              self.keys.fieldno = self.columns.fieldno
              self.keys.sortoptions = self.columns.sortoptions
              self.keys.isastring = self.columns.isastring
              add(self.keys)
            end
            break
          end
        end
        myPtr1 = myPtr2
      end
      if ~self.multikey
        loop while records(self.keys) > 1
          get(self.keys,records(self.keys))
          delete(self.keys)
        end
      end
    end
    if ~self.stickysort &= null
      self.stickysort = self.showsort()
    end
    self.ShowHeaders
    self.Sort


sdkLCS.ShowHeaders  procedure

myPtr   long

    code

    loop myPtr = 1 to records(self.columns)
      get(self.columns, myPtr)
      self.keys.column = self.columns.column
      get(self.keys, +self.keys.column)
      if errorcode()
        self.listfeq{PROPLIST:Header, self.columns.column} = self.columns.headertext
      elsif self.prependindicators
        self.listfeq{PROPLIST:Header, self.columns.column} = '[' & choose(self.keys.direction < 0,'-','+') & choose(records(self.keys)=1,'',clip(self.keys.level)) & ']' & left(self.columns.headertext)
      else
        self.listfeq{PROPLIST:Header, self.columns.column} = clip(self.columns.headertext) & '[' & choose(self.keys.direction < 0,'-','+') & choose(records(self.keys)=1,'',clip(self.keys.level)) & ']'
      end
    end


sdkLCS.ShowSort  procedure ! ,string

myPtr       long
myResult    string(1024)

    code

    sort(self.keys, +self.keys.level)
    loop myPtr = 1 to records(self.keys)
      get(self.keys, myPtr)
      if self.keys.direction < 0
        myResult = clip(myResult) & '-' & self.keys.column
      else
        myResult = clip(myResult) & '+' & self.keys.column
      end
    end
    return clip(myResult)


sdkLCS.Sort procedure

mySaveQ &string
myPtr   long

    code

    if self.listfrom &= null
      return
    end
    mySaveQ &= new string(size(self.listfrom))
    if records(self.listfrom) and records(self.keys)
      if self.showcursoronsort
        setcursor(CURSOR:Wait)
      end
      myPtr = choice(self.listfeq)
      if myPtr
        get(self.listfrom, myPtr)
        mySaveQ = self.listfrom
      end
      sort(self.keys, +self.keys.level)
      SDKLCSkeyqref &= self.keys
      SDKLCSSortFreeRTL(self.listfrom, SDKLCSCompareFunc)
      sort(self.listfrom, SDKLCSCompareFunc)
      if myPtr
        loop myPtr = 1 to records(self.listfrom)
          get(self.listfrom, myPtr)
          if mySaveQ = self.listfrom
            self.listfeq{PROP:Selected} = myPtr
            break
          end
        end
      end
      if self.showcursoronsort
        setcursor
      end
    end
    self.Storechanges = changes(self.listfrom)
    dispose(mySaveQ)


sdkLCS.TakeEvent  procedure

mysavesort  string(1024)
myptr       long

    code

    if self.listfrom &= null
      return
    end
    if self.sortonchanges
      if self.storechanges <> changes(self.listfrom)
        self.Sort
        return
      end
    end
    if self.locked
      return
    end
    if field() = self.listfeq
      case event()
      of EVENT:AlertKey
        case keycode()
        of self.alertkeystd orof self.alertkeymulti orof self.alertkeyclear
          if self.listfeq{PROPLIST:MouseDownRow} = 0 and self.listfeq{PROPLIST:MouseDownZone} = LISTZONE:header
            mysavesort = self.ShowSort()
            if keycode() = self.alertkeyclear
              free(self.keys)
            else
              sort(self.keys, +self.keys.level)
              self.columns.column = self.listfeq{PROPLIST:MouseDownField}
              get(self.columns, +self.columns.column)
              if ~errorcode() and ~band(self.columns.sortoptions,2^0)
                self.keys.column = self.columns.column
                get(self.keys, +self.keys.column)
                if errorcode()
                  if ~self.multikey or keycode() = self.alertkeystd
                    free(self.keys)
                  end
                  clear(self.keys)
                  self.keys.column = self.columns.column
                  self.keys.direction = 1
                  self.keys.fieldno = self.columns.fieldno
                  self.keys.sortoptions = self.columns.sortoptions
                  self.keys.isastring = self.columns.isastring
                  add(self.keys)
                else
                  self.keys.direction *= -1
                  if ~self.multikey or keycode() = self.alertkeystd
                    free(self.keys)
                    add(self.keys)
                  else
                    put(self.keys)
                  end
                end
                if (band(self.keys.sortoptions,2^2) and self.keys.direction < 0) or (band(self.keys.sortoptions,2^3) and self.keys.direction > 0)
                  self.keys.direction *= -1
                  put(self.keys)
                end
              end
            end
            loop myPtr = 1 to records(self.keys)
              get(self.keys, myPtr)
              self.keys.level = myPtr
              put(self.keys)
            end
            if ~self.stickysort &= null
              self.stickysort = self.showsort()
            end
            if mysavesort <> self.ShowSort()
              self.ShowHeaders
              self.Sort
            end
          end
        end
      end
    end


sdkLCS.Construct  procedure ! ,private

    code

    self.columns &= new sdkLCScolq
    self.keys &= new sdkLCSkeyq


sdkLCS.Destruct  procedure ! ,private

    code

    dispose(self.columns)
    dispose(self.keys)


SDKLCSCompareFunc procedure(*group cfTarget, *group cfCompare) ! ,signed

myPtr       long
myanyt      any
myanyc      any

    code

    loop myPtr = 1 to records(SDKLCSkeyqref)
      get(SDKLCSkeyqref, myPtr)
      myanyt &= what(cfTarget, SDKLCSkeyqref.fieldno)
      myanyc &= what(cfCompare, SDKLCSkeyqref.fieldno)
      if SDKLCSkeyqref.isastring and ~band(SDKLCSkeyqref.sortoptions,2^1)
        ! case-less string sort required
        if upper(myanyt) < upper(myanyc)
          return -SDKLCSkeyqref.direction
        elsif upper(myanyt) > upper(myanyc)
          return SDKLCSkeyqref.direction
        end
      elsif myanyt < myanyc
        return -SDKLCSkeyqref.direction
      elsif myanyt > myanyc
        return SDKLCSkeyqref.direction
      end
    end
    return 0 ! all keys are equal

